<?php

namespace FiloBlu\Flow\Model\Indexer;

use Exception;
use Magento\Catalog\Model\Product;
use Magento\CatalogInventory\Model\Stock\Status;
use Magento\Framework\Api\SearchCriteriaBuilder;
use Magento\Framework\App\ResourceConnection;
use Magento\Framework\DB\Adapter\AdapterInterface;
use Magento\Framework\Exception\LocalizedException;
use Magento\Framework\Indexer\ActionInterface;
use Magento\Framework\Mview\ActionInterface as MviewActionInterface;
use Magento\Framework\ObjectManagerInterface;
use Magento\Inventory\Model\ResourceModel\StockSourceLink\Collection;
use Magento\Inventory\Model\StockSourceLink;
use Magento\InventoryApi\Api\Data\SourceItemInterface;
use Magento\InventoryApi\Api\SourceItemRepositoryInterface;
use Magento\InventoryApi\Api\SourceItemsSaveInterface;
use Magento\InventoryConfigurationApi\Api\GetStockItemConfigurationInterface;
use Magento\InventoryConfigurationApi\Exception\SkuIsNotAssignedToStockException;
use Magento\InventoryReservationsApi\Model\GetReservationsQuantityInterface;
use Psr\Log\LoggerInterface;
use Throwable;

/**
 *
 */
class Stock implements ActionInterface, MviewActionInterface
{
    const INDEXER_ID = 'filoblu_flow_stock';

    /**
     * @var SourceItemRepositoryInterface
     */
    private $sourceItemRepository;

    /**
     * @var SearchCriteriaBuilder
     */
    private $searchCriteriaBuilder;

    /**
     * @var ResourceConnection
     */
    private $resourceConnection;

    /**
     * @var Product
     */
    private $product;

    /**
     * @var LoggerInterface
     */
    private $logger;

    /**
     * @var GetReservationsQuantityInterface
     */
    private $getReservationsQuantity;

    /**
     * @var GetStockItemConfigurationInterface
     */
    private $getStockItemConfiguration;

    /**
     * @var AdapterInterface
     */
    private $connection;

    /**
     * @var Collection
     */
    private $sourceCollection;

    /**
     * @var array|null
     */
    private $sourceToStockTable;

    /**
     * @var  array|null
     */
    private $sourcesToStockId;

    /**
     * @var array|null
     */
    private $stockIdToSources;

    /**
     * @param SearchCriteriaBuilder $searchCriteriaBuilder
     * @param ObjectManagerInterface $objectManager
     * @param ResourceConnection $resourceConnection
     * @param Product $product
     * @param LoggerInterface $logger
     * @param GetReservationsQuantityInterface $getReservationsQuantity
     * @param GetStockItemConfigurationInterface $getStockItemConfiguration
     */
    public function __construct(
        SearchCriteriaBuilder              $searchCriteriaBuilder,
        ObjectManagerInterface             $objectManager,
        ResourceConnection                 $resourceConnection,
        Product                            $product,
        LoggerInterface                    $logger,
        GetReservationsQuantityInterface   $getReservationsQuantity,
        GetStockItemConfigurationInterface $getStockItemConfiguration
    )
    {
        $this->searchCriteriaBuilder = $searchCriteriaBuilder;

        if (interface_exists(SourceItemRepositoryInterface::class)) {
            $this->sourceItemRepository = $objectManager->get(SourceItemRepositoryInterface::class);
        }

        if (class_exists(Collection::class)) {
            $this->sourceCollection = $objectManager->get(Collection::class);
        }

        $this->resourceConnection = $resourceConnection;
        $this->product = $product;
        $this->logger = $logger;
        $this->getReservationsQuantity = $getReservationsQuantity;
        $this->getStockItemConfiguration = $getStockItemConfiguration;
    }

    /**
     *
     */
    public function executeFull()
    {
        $this->execute(null);
    }

    /**
     * @param int[] $ids
     */
    public function execute($ids)
    {
        try {

            if (!$this->sourceItemRepository) {
                return;
            }

            if ($ids) {
                $this->searchCriteriaBuilder->addFilter('source_item_id', $ids, 'in');
            }

            $result = $this->sourceItemRepository->getList($this->searchCriteriaBuilder->create());

            foreach ($result->getItems() as $sourceItem) {
                //
                // @see \Magento\InventoryConfiguration\Model\GetLegacyStockItem::execute
                // Stock::DEFAULT_STOCK_ID is used until we have proper multi-stock item configuration
                //
                $stockItemConfiguration = $this->getStockItemConfiguration->execute(
                    $sourceItem->getSku(),
                    \Magento\CatalogInventory\Model\Stock::DEFAULT_STOCK_ID

                );

                // We don't need to Manage Stock
                if (!$stockItemConfiguration->isManageStock() || $stockItemConfiguration->getBackorders()) {
                    continue;
                }

                $this->updateLegacyStockFromSourceItem($sourceItem);
            }

        } catch (Exception $e) {
            $this->logger->critical('FiloBlu_Flow_Stock: ' . $e->getMessage());
        } catch (Throwable $t) {
            $this->logger->critical('FiloBlu_Flow_Stock: ' . $t->getMessage());
        }
    }

    /**
     * @param SourceItemInterface $sourceItem
     * @return mixed
     * @throws LocalizedException
     * @throws SkuIsNotAssignedToStockException
     */
    public function updateLegacyStockFromSourceItem(SourceItemInterface $sourceItem)
    {
        $sourceCode = $sourceItem->getSourceCode();
        $sku = $sourceItem->getSku();
        $stockId = $this->getStockId($sourceCode);
        $quantity = 0;
        $stockStatus = SourceItemInterface::STATUS_OUT_OF_STOCK;
        $stockQuantity = $this->getQuantity($sku, $stockId);
        $productId = $this->product->getIdBySku($sourceItem->getSku());

        $reservations = $this->getReservationsQuantity->execute($sourceItem->getSku(), $stockId);
        $minimumQuantity = $this->getStockItemConfiguration->execute($sourceItem->getSku(), $stockId)->getMinQty();

        $realStockStatus = (int)($stockQuantity + $reservations - $minimumQuantity) > 0 ?
            SourceItemInterface::STATUS_IN_STOCK : SourceItemInterface::STATUS_OUT_OF_STOCK;

        if ($realStockStatus != SourceItemInterface::STATUS_OUT_OF_STOCK) {
            $quantity = $stockQuantity;
            $stockStatus = SourceItemInterface::STATUS_IN_STOCK;
        }

        // For default legacy stock
        if ($sourceItem->getSourceCode() === 'default') {
            $this->updateStockItem($productId, $quantity, $stockStatus);
            $this->updateStockStatus($productId, $quantity, $stockStatus);
            return $productId;
        }

        // For sources != default
        $this->getConnection()->update(
            $this->getStockTable($sourceCode),
            [
                'is_salable' => $stockStatus,
                'quantity'   => $quantity
            ],
            ['sku = ?' => $sku]
        );

        return $productId;
    }

    /**
     * @param string $sourceCode
     * @return int
     */
    public function getStockId(string $sourceCode): int
    {
        if ($this->sourcesToStockId === null) {
            $this->createLookupTables();
        }

        return (int)$this->sourcesToStockId[$sourceCode];
    }

    /**
     *
     */
    public function createLookupTables()
    {
        $this->sourceToStockTable = [];
        $this->sourcesToStockId = [];
        $this->stockIdToSources = [];

        /** @var StockSourceLink $sourcesToStock */
        foreach ($this->sourceCollection->getItems() as $sourcesToStock) {

            $this->sourcesToStockId[$sourcesToStock->getSourceCode()] = $sourcesToStock->getStockId();

            $this->sourceToStockTable[$sourcesToStock->getSourceCode()] = $this->getConnection()
                ->getTableName("inventory_stock_{$sourcesToStock->getStockId()}");

            $this->stockIdToSources[$sourcesToStock->getStockId()][$sourcesToStock->getSourceCode()] =
                $sourcesToStock->getSourceCode();
        }
    }

    /**
     * @return AdapterInterface
     */
    public function getConnection(): AdapterInterface
    {
        if ($this->connection === null) {
            $this->connection = $this->resourceConnection->getConnection();
        }

        return $this->connection;
    }

    /**
     * @param $sku
     * @param $stockId
     * @return float
     */
    public function getQuantity($sku, $stockId): float
    {
        $sources = implode(
            ',',
            array_map(function ($a) {
                return "'{$a}'";
            }, $this->getSourcesFromStockId($stockId))
        );

        return (float)$this->getConnection()
            ->fetchOne("SELECT COALESCE(SUM(quantity),0) AS `quantity` FROM `inventory_source_item` WHERE `sku` = '{$sku}' AND `source_code` IN ({$sources});");
    }

    /**
     * @param $stockId
     * @return mixed
     */
    public function getSourcesFromStockId($stockId)
    {
        if ($this->stockIdToSources === null) {
            $this->createLookupTables();
        }

        return $this->stockIdToSources[$stockId];
    }

    /**
     * @param $productId
     * @param $quantity
     * @param $isInStock
     */
    public function updateStockItem($productId, $quantity, $isInStock)
    {
        $connection = $this->getConnection();
        $tableName = $connection->getTableName('cataloginventory_stock_item');
        $connection->update($tableName,
            [
                'is_in_stock' => $isInStock,
                'qty'         => $quantity
            ],
            ['product_id = ?' => $productId]
        );
    }

    /**
     * @param int $productId
     * @param int $isInStock
     * @param int $quantity
     */
    public function updateStockStatus(int $productId, int $quantity, int $isInStock)
    {
        $connection = $this->getConnection();
        $tableName = $connection->getTableName('cataloginventory_stock_status');
        $connection->update($tableName, [
            'stock_status' => $isInStock,
            'qty'          => $quantity
        ],
            ['product_id = ?' => $productId]
        );
    }

    /**
     * @param string $sourceCode
     * @return string
     */
    public function getStockTable(string $sourceCode): string
    {
        if ($this->sourceToStockTable === null) {
            $this->createLookupTables();
        }

        return $this->sourceToStockTable[$sourceCode];
    }

    /**
     * @param array $ids
     */
    public function executeList(array $ids)
    {
        $this->execute($ids);
    }

    /**
     * @param int $id
     */
    public function executeRow($id)
    {
        $this->execute([$id]);
    }
}
